//
// Created by nicolas on 22/03/18.
//

// usleep
#ifdef _WIN32
#include <windows.h>
#define usleep Sleep
#endif
#ifdef __linux__
#include <unistd.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "connection.h"
#include "subscription.h"
#include "tcp_package.h"
#include "tcp_messages.h"
#include "credentials.h"
#include "proto_helper.h"
#include "utils/array.h"
#include "utils/debug.h"
#include "utils/error.h"
#include "utils/string.h"
#include "utils/socket.h"
#include "proto.h"

struct st_esc_connection_settings default_connection_settings = {
    BOOL_FALSE
};

esc_connection_settings_t* esc_default_connection_settings = &default_connection_settings;

//TODO partial transfer
ssize_t connection_send_tcp_package(esc_connection_t* conn, tcp_package_t* pkg) {
    char uuid_buf[37];
#ifdef _WIN32
    fprintf(stderr, "connection_send_tcp_package: %s %u %s %llu ", get_string_for_tcp_message(pkg->command), pkg->flags, uuid_format(pkg->correlation_id, uuid_buf, 37), buffer_size(pkg->data));
#else
    fprintf(stderr, "connection_send_tcp_package: %s %u %s %lu ", get_string_for_tcp_message(pkg->command), pkg->flags,
            uuid_format(pkg->correlation_id, uuid_buf, 37), buffer_size(pkg->data));
#endif
    buffer_t* send_buffer = tcp_package_to_buffer(pkg);

    size_t send_buffer_size = buffer_size(send_buffer);
    uint32_t size = (uint32_t)send_buffer_size;
    ssize_t rc;
    if ((rc = socket_send(conn->tcp_conn, (char *) &size, sizeof(uint32_t))) <= 0) {
        buffer_destroy(send_buffer);
        fprintf(stderr, "%d\n", socket_error(conn->tcp_conn));
        return -1;
    }

    uint8_t* send_buffer_data = buffer_data(send_buffer);
    if ((rc = socket_send(conn->tcp_conn, (char *) send_buffer_data, send_buffer_size)) <= 0) {
        buffer_destroy(send_buffer);
        fprintf(stderr, "%d\n", socket_error(conn->tcp_conn));
        return -2;
    }

    buffer_destroy(send_buffer);
    fprintf(stderr, "0\n");
    return 0;
}

//TODO partial transfer
tcp_package_t* connection_recv_tcp_package(esc_connection_t* conn) {
    uint32_t recv_size;
    ssize_t rc;
    if ((rc = socket_recv(conn->tcp_conn, (char *)&recv_size, sizeof(uint32_t))) != 4) {
#ifdef _WIN32
        fprintf(stderr, "connection_recv_tcp_package: %lld %d\n", rc, socket_error(conn->tcp_conn));
#else
        fprintf(stderr, "connection_recv_tcp_package: %ld %d\n", rc, socket_error(conn->tcp_conn));
#endif
        return 0;
    }
    buffer_t* recv_buffer = buffer_create(recv_size);
    size_t recv_buffer_size = recv_size;
    uint8_t* recv_buffer_data = buffer_data(recv_buffer);
    while(recv_size > 0) {
        size_t pos = recv_buffer_size - recv_size;
        rc = socket_recv(conn->tcp_conn, (char *)&recv_buffer_data[pos], recv_size);
        recv_size -= rc;
    }
    tcp_package_t* recv_pkg = tcp_package_from_buffer(recv_buffer);
    buffer_destroy(recv_buffer);
    char uuid_buf[37];
#ifdef _WIN32
    fprintf(stderr, "connection_recv_tcp_package: %s %u %s %llu\n", get_string_for_tcp_message(recv_pkg->command), recv_pkg->flags, uuid_format(recv_pkg->correlation_id, uuid_buf, 37), buffer_size(recv_pkg->data));
#else
    fprintf(stderr, "connection_recv_tcp_package: %s %u %s %lu\n", get_string_for_tcp_message(recv_pkg->command), recv_pkg->flags,
            uuid_format(recv_pkg->correlation_id, uuid_buf, 37), buffer_size(recv_pkg->data));
#endif
    return recv_pkg;
}

void* connection_thread(void* arg) {
    esc_connection_t* conn = arg;

    while(conn->stop == BOOL_FALSE) {
        if (socket_writable(conn->tcp_conn)) {
            tcp_package_t* send_pkg = queue_dequeue(conn->send_queue);
            if (send_pkg != 0) {
                ssize_t rc = connection_send_tcp_package(conn, send_pkg);
                tcp_package_destroy(send_pkg);
                if (rc != 0) {
                    fprintf(stderr, "failed to send pkg.");
                }
            }
        }

        if (socket_readable(conn->tcp_conn)) {
            esc_subscription_t* subscription = 0;
            tcp_package_t* recv_pkg = connection_recv_tcp_package(conn);
            if (recv_pkg == 0) {
                //TODO Handle connection lost
                conn->stop = 1;
            } else if (recv_pkg->command == MESSAGE_HEARTBEATREQUEST) {
                tcp_package_t* heartbeat_pkg = tcp_package_create(MESSAGE_HEARTBEATRESPONSE, recv_pkg->correlation_id, buffer_create(0));
                queue_enqueue(conn->send_queue, heartbeat_pkg);
                tcp_package_destroy(recv_pkg);
            } else if ((subscription = queue_find(conn->subscriptions, esc_find_subscription_by_correlation_id, recv_pkg->correlation_id)) != 0) {
                esc_subscription_handle_tcp_package(subscription, recv_pkg);
            } else {
                queue_enqueue(conn->recv_queue, recv_pkg);
            }
        }

        usleep(1);
    }

    printf("manager thread stopped\n");
    event_set(conn->stopped_event);
}

bool_t find_tcp_package_by_correlation_id(void* _item, void* _correlation_id) {
    tcp_package_t* item = _item;
    uuid_t* correlation_id = _correlation_id;
    if (uuid_compare(item->correlation_id, correlation_id) == 0) {
        return BOOL_TRUE;
    }
    return BOOL_FALSE;
}

//TODO add timeout
tcp_package_t* connection_wait_for(esc_connection_t* conn, uuid_t* correlation_id) {
    tcp_package_t* found = 0;
    while(found == 0) {
        found = queue_remove(conn->recv_queue, find_tcp_package_by_correlation_id, correlation_id);
        usleep(1);
    }
    return found;
}

esc_connection_t* esc_connection_create(esc_connection_settings_t* connection_settings, const char* addr, const char* connection_name) {
    struct st_esc_connection* conn = malloc(sizeof(struct st_esc_connection));
    conn->settings = connection_settings;
    if (connection_name == 0 || strcmp(connection_name, "") == 0) {
        uuid_t* uuid = uuid_create();
        char buf[40];
        uuid_format(uuid, buf, 40);
        conn->name = string_copy(buf);
        uuid_destroy(uuid);
    } else {
        conn->name = connection_name;
    }
    conn->send_queue = queue_create();
    conn->recv_queue = queue_create();
    conn->stop = 0;
    conn->stopped_event = event_create();
    conn->last_error = 0;
    conn->subscriptions = queue_create();

    if (strncmp(addr, "tcp://", 6) != 0) {
        conn->last_error = error_create(1, "invalid schema for address: %s", addr);
        return conn;
    }
    char* pos = strrchr(addr, ':');
    if (pos == 0) {
        conn->last_error = error_create(2, "missing port in address: %s", addr);
        return conn;
    }
    char* host = malloc(pos - addr - 5);
    strncpy(host, addr+6, pos-addr-6);
    host[pos-addr-6] = 0;
    unsigned short port = (unsigned short)atoi(pos+1);
    esc_static_endpoint_discoverer_t discover_data = {
            {host, port}, connection_settings->use_ssl_connection
    };

    conn->discoverer_data = malloc(sizeof(struct st_esc_static_endpoint_discoverer));
    memcpy(conn->discoverer_data, &discover_data, sizeof(struct st_esc_static_endpoint_discoverer));
    conn->discover = (esc_endpoint_discoverer_t)esc_static_discover;

    return conn;
}

void esc_connection_destroy(esc_connection_t* conn) {
    esc_connection_close(conn);

    socket_destroy(conn->tcp_conn);
    free(((struct st_esc_static_endpoint_discoverer*)conn->discoverer_data)->tcp_endpoint.host);
    free(conn->discoverer_data);
    thread_destroy(conn->manager_thread);
    queue_destroy(conn->recv_queue);
    queue_destroy(conn->send_queue);
    event_destroy(conn->stopped_event);
    free((void*)conn->name);
    queue_destroy(conn->subscriptions);
    free(conn);
}

// return 0 on success
// return non-zero on failure and sets last_error on connection
int esc_connection_connect(esc_connection_t* conn) {
    // Discover endpoint
    esc_node_endpoints_t* endpoints = conn->discover(conn->discoverer_data, 0);
    if (endpoints == 0) {
        return -1;
    }
    // Establish Tcp Connection
    conn->tcp_conn = socket_create(SOCKET_TYPE_TCP);
    ip_endpoint_t endpoint = conn->settings->use_ssl_connection ? endpoints->secure_tcp_endpoint : endpoints->tcp_endpoint;
    if (socket_connect(conn->tcp_conn, endpoint.host, endpoint.port) != 0) {
        conn->last_error = error_create(socket_error(conn->tcp_conn), "can't connect to %s:%d", endpoint.host, endpoint.port);
        return -1;
    }
    free(endpoints);
    // Start manager thread
    conn->manager_thread = thread_create(connection_thread, conn);
    if (thread_start(conn->manager_thread)) {
        conn->last_error = error_create(thread_error(conn->manager_thread), "can't start manager thread.", "");
        return -1;
    }
    // Identify
    uuid_t* correlation_id = uuid_create();

    buffer_t* msg_buf = esc_identify_client_pack(conn->name);
    tcp_package_t* send_pkg = tcp_package_create(MESSAGE_IDENTIFYCLIENT, correlation_id, msg_buf);
    queue_enqueue(conn->send_queue, send_pkg);

    tcp_package_t* recv_pkg = connection_wait_for(conn, correlation_id);
    if (recv_pkg->command != MESSAGE_CLIENTIDENTIFIED) {
        conn->last_error = error_create(0, "server error: %d.", recv_pkg->command);
        uuid_destroy(correlation_id);
        tcp_package_destroy(recv_pkg);
        return -1;
    }

    uuid_destroy(correlation_id);
    tcp_package_destroy(recv_pkg);
    return 0;
}

esc_all_events_slice_t* esc_connection_read_all_forward(esc_connection_t* conn, esc_position_t* last_checkpoint, unsigned int count, esc_credentials_t* credentials) {
    uuid_t* correlation_id = uuid_create();

    buffer_t* msg_buf = esc_read_all_forward_pack(last_checkpoint, count, 0);
    tcp_package_t* send_pkg = tcp_package_create_authenticated(MESSAGE_READALLEVENTSFORWARD, correlation_id, msg_buf, credentials->username, credentials->password);
    queue_enqueue(conn->send_queue, send_pkg);

    tcp_package_t* recv_pkg = connection_wait_for(conn, correlation_id);
    if (recv_pkg->command != MESSAGE_READALLEVENTSFORWARDCOMPLETED) {
        conn->last_error = error_create(recv_pkg->command, "wrong message from the server: %s", get_string_for_tcp_message(recv_pkg->command));
        tcp_package_destroy(recv_pkg);
        uuid_destroy(correlation_id);
        return 0;
    }

    esc_all_events_slice_t* result;
    inspection_result_t* inspection_result = esc_all_events_slice_unpack(recv_pkg->data, &result);
    // TODO fully handle inspection results
    if (inspection_result->error) {
        conn->last_error = error_copy(inspection_result->error);
        inspection_result_destroy(inspection_result);
        tcp_package_destroy(recv_pkg);
        uuid_destroy(correlation_id);
        return 0;
    }
    inspection_result_destroy(inspection_result);
    tcp_package_destroy(recv_pkg);
    uuid_destroy(correlation_id);
    return result;
}

esc_write_result_t* esc_append_to_stream(esc_connection_t* conn, const char* stream, int64_t expected_version, array_t* events) {
    uuid_t* correlation_id = uuid_create();

    buffer_t* msg_buf = esc_append_to_stream_pack(stream, expected_version, events);
    tcp_package_t* send_pkg = tcp_package_create(MESSAGE_WRITEEVENTS, correlation_id, msg_buf);
    queue_enqueue(conn->send_queue, send_pkg);

    tcp_package_t* recv_pkg = connection_wait_for(conn, correlation_id);
    if (recv_pkg->command != MESSAGE_WRITEEVENTSCOMPLETED) {
        conn->last_error = error_create(recv_pkg->command, "server error: %s", get_string_for_tcp_message(recv_pkg->command));
        tcp_package_destroy(recv_pkg);
        uuid_destroy(correlation_id);
        return 0;
    }

    esc_write_result_t* result;
    inspection_result_t* inspection_result = esc_write_result_unpack(recv_pkg->data, &result);
    // TODO fully handle inspection results
    if (inspection_result->error) {
        conn->last_error = error_copy(inspection_result->error);
        inspection_result_destroy(inspection_result);
        tcp_package_destroy(recv_pkg);
        uuid_destroy(correlation_id);
        return 0;
    }
    inspection_result_destroy(inspection_result);
    tcp_package_destroy(recv_pkg);
    uuid_destroy(correlation_id);
    return result;
}

esc_subscription_t* esc_connection_subscribe_to_all(esc_connection_t* conn, bool_t resolve_link_tos, esc_event_appeared_callback_t event_appeared_callback, esc_subscription_dropped_callback_t subscription_dropped_callback, esc_credentials_t* credentials) {
    uuid_t* correlation_id = uuid_create();

    esc_subscription_t* subscription = esc_subscription_create(conn, correlation_id, event_appeared_callback, subscription_dropped_callback);
    queue_enqueue(conn->subscriptions, subscription);

    buffer_t* msg_buf = esc_subscribe_to_stream_pack("", resolve_link_tos);
    tcp_package_t* send_pkg = tcp_package_create_authenticated(MESSAGE_SUBSCRIBETOSTREAM, correlation_id, msg_buf, credentials->username, credentials->password);
    queue_enqueue(conn->send_queue, send_pkg);

    bool_t subscribing = BOOL_TRUE;
    while(subscribing) {
        subscribing = esc_subscription_state(subscription) == ESC_SUBSCRIBING;
        usleep(1);
    }

    uuid_destroy(correlation_id);
    return subscription;
}

void esc_connection_close(esc_connection_t* conn) {
    if (conn->stop == 1) return;
    conn->stop = 1;
    event_wait(conn->stopped_event);
    socket_close(conn->tcp_conn);
}

inline error_t* esc_connection_last_error(esc_connection_t* conn) {
    return conn->last_error;
}
